// contactmodelBeam.cpp
#include "contactmodelBeam.h"

#include "module/interface/icontactmechanical.h"
#include "module/interface/icontact.h"
#include "module/interface/ipiecemechanical.h"
#include "module/interface/ipiece.h"
#include "module/interface/ifishcalllist.h"

#include "version.txt"

#include "utility/src/tptr.h"
#include "base/src/mathutil.h"

#include "kernel/interface/iprogram.h"
#include "module/interface/icontactthermal.h"
#include "contactmodel/src/contactmodelthermal.h"

#ifdef Beam_LIB
    int __stdcall DllMain(void *,unsigned, void *) {
        return 1;
    }

    extern "C" EXPORT_TAG const char *getName() {
#if DIM==3
        return "contactmodelmechanical3dBeam";
#else
        return "contactmodelmechanical2dBeam";
#endif
    }

    extern "C" EXPORT_TAG unsigned getMajorVersion() {
        return MAJOR_VERSION;
    }

    extern "C" EXPORT_TAG unsigned getMinorVersion() {
        return MINOR_VERSION;
    }

    extern "C" EXPORT_TAG void *createInstance() {
        cmodelsxd::ContactModelBeam *m = new cmodelsxd::ContactModelBeam();
        return (void *)m;
    }
#endif 

namespace cmodelsxd {
    using namespace itasca;

    int ContactModelBeam::index_ = -1;
    UInt ContactModelBeam::getMinorVersion() const { return MINOR_VERSION;}

    ContactModelBeam::ContactModelBeam() : E_(0.0)
		                                 , nu_(0.0)
										 , bhGiven_(0)
		                                 , A_(0.0)
		                                 , J_(0.0)
		                                 , Iy_(0.0)
		                                 , Iz_(0.0)
										 , b_(0.0)
										 , h_(0.0)
		                                 , Ydir_( DVect3(0.0,1.0,0.0) )
		                                 , L_(1.0)
		                                 , kn_(0.0)
		                                 , ks_(0.0)
		                                 , F_( DVect3(0.0) )
		                                 , M_( DVect3(0.0) )
										 , Estr_(0.0)
	                                     , cosAl_(1.0)
	                                     , sinAl_(0.0)
	                                     , cycled_(false)
	                                     , Lset_(false) {
	}

    ContactModelBeam::~ContactModelBeam() {
    }

    void ContactModelBeam::archive(ArchiveStream &stream) {
        // The stream allows one to archive the values of the contact model
        // so that it can be saved and restored. The minor version can be
        // used here to allow for incremental changes to the contact model too.
        stream & E_;
        stream & nu_;
		stream & bhGiven_;
        stream & A_;
        stream & J_;
        stream & Iy_;
        stream & Iz_;
		stream & b_;
		stream & h_;
        stream & Ydir_;
        stream & L_;
        stream & kn_;
        stream & ks_;
        stream & F_;
        stream & M_;
		stream & Estr_;
		//
		stream & cosAl_;
        stream & sinAl_;
		stream & cycled_;
		stream & Lset_;
    }

    void ContactModelBeam::copy(const ContactModel *cm) {
        // Copy all of the contact model properties. Used in the CMAT 
        // when a new contact is created. 
        ContactModelMechanical::copy(cm);
        const ContactModelBeam *in = dynamic_cast<const ContactModelBeam*>(cm);
        if (!in) throw std::runtime_error("Internal error: contact model dynamic cast failed.");
		E_    = in->E_;
		nu_   = in->nu_;
		bhGiven_ = in->bhGiven_;
		A_    = in->A_;
		J_    = in->J_;
		Iy_   = in->Iy_;
		Iz_   = in->Iz_;
		b_    = in->b_;
		h_    = in->h_;
		Ydir_ = in->Ydir_;
		L_    = in->L_;
		kn_   = in->kn_;
		ks_   = in->ks_;
		F_    = in->F_;
		M_    = in->M_;
		Estr_ = in->Estr_;
		//
		cosAl_  = in->cosAl_;
		sinAl_  = in->sinAl_;
		cycled_ = in->cycled_;
		Lset_   = in->Lset_;
	}

    QVariant ContactModelBeam::getProperty(uint i,const IContact *con) const {
      // Return the property. The IContact pointer is provided so that 
      // more complicated properties, depending on contact characteristics,
      // can be calculated. 
      QVariant var;
      switch (i) {
        case kwE:    return E_;
        case kwNu:   return nu_;
		case kwBHgiven: return bhGiven_;
        case kwA:    return A_;
        case kwJ:    return J_;
        case kwIy:   return Iy_;
        case kwIz:   return Iz_;
        case kwb:    return b_;
        case kwh:    return h_;
        case kwYdir: var.setValue( Ydir_ ); return var;
        case kwL:    return L_;
        case kwKn:   return kn_;
        case kwKs:   return ks_;
		case kwYaxis: {
		  DVect3 y_b(0.0);  
          if (cycled_) {  // Beam coord. system has been set and frozen.
			// Sst = [tgl].transpose * Syz, set Syz = |_ 1 0 _]transpose
			// Sst = |_ cosAl_ sinAl_ _|transpose
			CAxes localSys = con->getLocalSystem();
            DVect3 sc = localSys.e2();
            DVect3 tc = localSys.e3();
			y_b = sc * cosAl_  +  tc * sinAl_;
			y_b = y_b.unit();
		  }
		  var.setValue( y_b );
		  return var;
		}
        case kwForce:  var.setValue( F_ ); return var;
        case kwMoment: var.setValue( M_ ); return var;
		case kwEstr:   return Estr_;
      }
      assert(0);
      return QVariant();
    }

    bool ContactModelBeam::getPropertyGlobal(uint i) const {
      // Returns whether or not a property is held in the global axis system (TRUE)
      // or the local system (FALSE). Used by the plotting logic.
      switch (i) {
        case kwForce:  return false;   
        case kwMoment: return false;  
      }
      return true;
    }

    bool ContactModelBeam::setProperty( uint i, const QVariant &v, IContact * ) {
      // Set a contact model property. Return value indicates that the timestep
      // should be recalculated. 
      switch (i) { 
        case kwE: {
          if ( !v.canConvert<double>() )
            throw Exception("bm_E must be a double.");
          double val( v.toDouble() );
          if ( val < 0.0 )
            throw Exception("Negative bm_E not allowed.");
          E_ = val;
          return true;
        }
        case kwNu: {
          if ( !v.canConvert<double>() )
            throw Exception("bm_nu must be a double.");
          double val( v.toDouble() );
          if ( (val <= -1.0) || (val > 0.5) )
            throw Exception("bm_nu must be in range (-1.0, 0.5].");
          nu_ = val;
          return true;
        }
        case kwBHgiven: {
          if ( !v.canConvert<int>() )
            throw Exception("bm_bhGiven must be an integer.");
          int val( v.toInt() );
          if ( (val < 0) || (val > 1) )
			  throw Exception("bm_bhGiven must be {0,1}.");
          bhGiven_ = val;
          return true;
        }
        case kwA: {
          if ( !v.canConvert<double>() )
            throw Exception("bm_A must be a double.");
          double val( v.toDouble() );
          if ( val < 0.0 )
            throw Exception("Negative bm_A not allowed.");
          A_ = val;
          return true;
        }
        case kwJ: {
          if ( !v.canConvert<double>() )
            throw Exception("bm_J must be a double.");
          double val( v.toDouble() );
          if ( val < 0.0 )
            throw Exception("Negative bm_J not allowed.");
          J_ = val;
          return true;
        }
        case kwIy: {
          if ( !v.canConvert<double>() )
            throw Exception("bm_Iy must be a double.");
          double val( v.toDouble() );
          if ( val < 0.0 )
            throw Exception("Negative bm_Iy not allowed.");
          Iy_ = val;
          return true;
        }
        case kwIz: {
          if ( !v.canConvert<double>() )
            throw Exception("bm_Iz must be a double.");
          double val( v.toDouble() );
          if ( val < 0.0 )
            throw Exception("Negative bm_Iz not allowed.");
          Iz_ = val;
          return true;
        }
        case kwb: {
          if ( !v.canConvert<double>() )
            throw Exception("bm_b must be a double.");
          double val( v.toDouble() );
          if ( val < 0.0 )
            throw Exception("Negative bm_b not allowed.");
          b_ = val;
          return true;
        }
        case kwh: {
          if ( !v.canConvert<double>() )
            throw Exception("bm_h must be a double.");
          double val( v.toDouble() );
          if ( val < 0.0 )
            throw Exception("Negative bm_h not allowed.");
          h_ = val;
          return true;
        }
        case kwYdir: {
          if ( !v.canConvert<DVect3>() )
            throw Exception("bm_Ydir must be a 3-component vector.");
          DVect3 val( v.value<DVect3>() );
          Ydir_ = val;
          return false;
        }
      }
      return false;
    }

    bool ContactModelBeam::getPropertyReadOnly(uint i) const {
      // Returns TRUE if a property is read only or FALSE otherwise. 
      switch (i) {
        case kwL:
        case kwKn:
        case kwKs:
        case kwYaxis:
        case kwForce:
        case kwMoment:
		case kwEstr:
          return true;
      }
      return false;
    }

    bool ContactModelBeam::validate( ContactModelMechanicalState *state, const double & ) {
      assert(state);

	  if (!Lset_) { // Do this only once.
	    setL( state->getMechanicalContact() );
		Lset_ = true;
	  }
	  kn_ = E_ / L_;
	  ks_ = E_ / ( 2.0*(1.0 + nu_)*L_ );

	  if (bhGiven_ == 1) {
		A_ = b_*h_;
		J_ = b_*h_*(b_*b_ + h_*h_) / 12.0;
		Iy_ = b_*h_*h_*h_ / 12.0;
		Iz_ = h_*b_*b_*b_ / 12.0;
	  }
	  return true;
    }

    bool ContactModelBeam::forceDisplacementLaw(ContactModelMechanicalState *state,const double &timestep) {
      assert(state);

	  if (!cycled_) { // Do this only during the first cycle.
		defineBeamCoordSys( state->getMechanicalContact() );
		cycled_ = true;
	  }

      // (1) Update Fn (we are storing -Fn in the first component).
	  F_.rdof(0) -= kn_ * A_ * state->relativeTranslationalIncrement_.rdof(0);

      // (2) Update Fs.
	  F_.rdof(1) -= ks_ * A_ * state->relativeTranslationalIncrement_.rdof(1);
	  F_.rdof(2) -= ks_ * A_ * state->relativeTranslationalIncrement_.rdof(2);

	  // (3) Update Mt.
	  M_.rdof(0) -= ks_ * J_ * state->relativeAngularIncrement_.rdof(0);

	  // (4) Update Mb.
	  updateMb( state->relativeAngularIncrement_.rdof(1), state->relativeAngularIncrement_.rdof(2) );

	  // Update the strain energy.
	  updateEstr();

	  // Update the contact force and moment.
	  state->force_ = F_;
	  state->momentOn1_ = M_;
	  state->momentOn2_ = M_;
      state->getMechanicalContact()->updateResultingTorquesLocal( state->force_,
                                                                  &state->momentOn1_, &state->momentOn2_ );
	  return true;
	}

    DVect2 ContactModelBeam::getEffectiveTranslationalStiffness() const {
	  double kn_sq = kn_ * A_ ;
	  double ks_sq = ks_ * A_ ;
      return( DVect2( kn_sq, ks_sq ) );
	}

    DAVect ContactModelBeam::getEffectiveRotationalStiffness() const {
      double kt_sq = ks_ * J_ ;
      double kb_sq = kn_ * max( Iy_, Iz_ );
	  DAVect ret;
        	 ret.rx() = kt_sq;
             ret.ry() = kb_sq;
             ret.rz() = kb_sq;
	  return ret;
	}

    void ContactModelBeam::resetForcesAndMoments() {
      DVect3 dvZero(0.0);
	  F_ = dvZero;
	  M_ = dvZero;
	}

	void ContactModelBeam::updateMb(const double &Ddss, const double &Ddst) {
	  Matrix<double, 2,2> tgl;
	  tgl[0][0] =  cosAl_;
	  tgl[0][1] =  sinAl_;
	  tgl[1][0] = -sinAl_;
	  tgl[1][1] =  cosAl_;

	  Matrix<double, 2,2> knI;
	  knI[0][0] = -kn_ * Iy_;
	  knI[0][1] = 0.0;
	  knI[1][0] = 0.0;
	  knI[1][1] = -kn_ * Iz_;

	  Matrix<double, 2, 1> del_thetab;
	  del_thetab(0) = Ddss;
	  del_thetab(1) = Ddst;

	  Matrix<double, 2, 1> del_Mb;
	  del_Mb = tgl.transpose() * knI * tgl * del_thetab;

	  M_.rdof(1) += del_Mb(0);
	  M_.rdof(2) += del_Mb(1);
	}

	void ContactModelBeam::defineBeamCoordSys( const IContactMechanical *con ) {
      CAxes localSys = con->getContact()->getLocalSystem();
	  DVect3 nc = localSys.e1();
	  DVect3 Ydir = getYdir( nc );

	  DVect3 Ypar = Ydir - nc*(Ydir | nc);
	  DVect3 yb = Ypar.unit();
	  DVect3 sc = localSys.e2();
	  DVect3 tc = localSys.e3();
	  cosAl_ = yb | sc;
	  sinAl_ = yb | tc;
	}

	DVect3 ContactModelBeam::getYdir( const DVect3 &nc ) const {
      DVect3 myDir = Ydir_.unit();
	  DVect3 myCross = myDir & nc;
	  if ( myCross.mag() > 1.0e-6 ) return myDir ;

	  myDir = DVect3( 0.0, 1.0, 0.0 );
	  myCross = myDir & nc;
	  if ( myCross.mag() > 1.0e-6 ) return myDir ;

	  myDir = DVect3( 1.0, 0.0, 0.0 );
	  return myDir ;
	}

	void ContactModelBeam::setL( const IContactMechanical *con ) {
      double R1 = 1.0 / con->getEnd1Curvature().rdof(1);      
      double R2 = 1.0 / con->getEnd2Curvature().rdof(1);      
	  double gap = con->getGap();
	  L_ = R1 + R2 + gap;
	}

	void ContactModelBeam::updateEstr() {
	  Estr_ = 0.0;
      
	  Estr_ += F_.rdof(0)*F_.rdof(0) / (kn_ * A_);

	  DVect2 dvFs( F_.rdof(1), F_.rdof(2) );
	  Estr_ += dvFs.mag()*dvFs.mag() / (ks_ * A_);

	  Estr_ += M_.rdof(0)*M_.rdof(0) / (ks_ * J_);

	  Matrix<double, 2,2> tgl;
	  tgl[0][0] =  cosAl_;
	  tgl[0][1] =  sinAl_;
	  tgl[1][0] = -sinAl_;
	  tgl[1][1] =  cosAl_;
	  Matrix<double, 2, 1> Mbst;
	  Mbst(0) = M_.rdof(1);
	  Mbst(1) = M_.rdof(2);
	  Matrix<double, 2, 1> Mbyz;
	  Mbyz = tgl * Mbst;

	  Estr_ += Mbyz(0)*Mbyz(0) / (kn_ * Iy_);
	  Estr_ += Mbyz(1)*Mbyz(1) / (kn_ * Iz_);

	  Estr_ *= 0.5;  // Don't forget to take one half of these terms
	}

} // namespace cmodelsxd
// EoF